program Oper

  use sep
  use mem_mod
  use fbi_mod
  use tomo_mod
  use separate_mod

  implicit none

  integer                             :: nt, n1, n2, wfld_j, i, gpu_n, gpu_o
  integer                             :: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
  integer                             :: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec
  integer                             :: bc1, bc2, bc3, bc4, ntau, otau, itau
  real                                :: dt, d1, o1, d2, o2

  real                                :: eps, dist1, dist2
  real, dimension(:), allocatable     :: wavelet, res
  real, dimension(:,:), allocatable   :: vel0, vel1, b0, b1, b_grad, lb, ub
  real, dimension(:,:,:), allocatable :: data_obs, data_obs2, data_res, p0, p1, p_res, p_grad, twt

  integer                             :: stat, iter1, niter1, iter2, niter2, itry, ntry, counter

  double precision                    :: r0nl, r0, r1, r2, r3, step_size

  integer, dimension(8)               :: timer0, timer1
  real                                :: timer_sum=0.

  call sep_init()

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer0)
!---------------------------------------------------------------------------

  call from_history("n1", n1)
  call from_history("o1", o1)
  call from_history("d1", d1)
  call from_history("n2", n2)
  call from_history("o2", o2)
  call from_history("d2", d2)

  call from_param("ntau", ntau, 1)
  call from_param("otau", otau, 0)
  if(ntau<1-otau) call erexit("zero tau axis is not included")

  call from_param("nt", nt)
  call from_aux("wavelet", "d1", dt)

  call from_param("bc1", bc1)
  call from_param("bc2", bc2)
  call from_param("bc3", bc3)
  call from_param("bc4", bc4)
  call from_param("wfld_j", wfld_j)

  call from_param("sou_nx", sou_nx)
  call from_param("sou_ox", sou_ox)
  call from_param("sou_dx", sou_dx)
  call from_param("sou_z", sou_z)
  if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > n2) call erexit("sou axis values are not correct")
  if(sou_z < 1 .or. sou_z > n1) call erexit("sou_z value is not correct")

  call from_param("rec_nx", rec_nx)
  call from_param("rec_ox", rec_ox)
  call from_param("rec_dx", rec_dx)
  call from_param("rec_z", rec_z)
  if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > n2) call erexit("rec axis values are not correct")
  if(rec_z < 1 .or. rec_z > n1) call erexit("rec_z value is not correct")

  call from_param("gpu_n", gpu_n)
  call from_param("gpu_o", gpu_o)

  call from_param('niter1', niter1, 10)
  call from_param('niter2', niter2, 10)
  call from_param('ntry', ntry, 5)

  call from_param('dist1', dist1)
  call from_param('dist2', dist2)

  call from_param('eps', eps)

  call to_history("n1", n1)
  call to_history("o1", o1)
  call to_history("d1", d1)
  call to_history("n2", n2)
  call to_history("o2", o2)
  call to_history("d2", d2)

  ! Allocation
  call mem_alloc1d(wavelet, nt)
  call mem_alloc1d(res, 2)
  call mem_alloc2d(vel0, n1, n2)
  call mem_alloc2d(vel1, n1, n2)
  call mem_alloc2d(b0, n1, n2)
  call mem_alloc2d(b1, n1, n2)
  call mem_alloc2d(b_grad, n1, n2)
  call mem_alloc2d(lb, n1, n2)
  call mem_alloc2d(ub, n1, n2)
  call mem_alloc3d(data_res, nt, rec_nx, sou_nx)
  call mem_alloc3d(data_obs, nt, rec_nx, sou_nx)
  call mem_alloc3d(data_obs2, nt, rec_nx, sou_nx)
  call mem_alloc3d(p0, n1, n2, ntau)
  call mem_alloc3d(p1, n1, n2, ntau)
  call mem_alloc3d(p_res, n1, n2, ntau)
  call mem_alloc3d(p_grad, n1, n2, ntau)
  call mem_alloc3d(twt, n1, n2, ntau)

  call sep_read(vel0)
  call sep_read(wavelet, "wavelet")
  call sep_read(data_obs, "data")
  call sep_read(lb, "lb")
  call sep_read(ub, "ub")

  where(vel0<lb) vel0=lb
  where(vel0>ub) vel0=ub

  ! Call L_initiate
  call init1(n1, o1, d1, n2, o2, d2, bc1, bc2, bc3, bc4)
  call init2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z)
  call init3(nt, dt, wfld_j, ntau, otau, gpu_n, gpu_o)
  call init_alloc(wavelet)
  call separate_init(n1, n2, d1, d2, dist1, dist2)

  do itau=1,ntau
    twt(:,:,itau) = itau-1+otau
  end do
  twt = abs(twt)
  if(maxval(twt)>0.) twt = eps*twt/maxval(twt)

  counter = 0

  do iter1=1,niter1
    call model_update(vel0)
    call L_nlop2(.false., data_res)
    data_obs2 = data_obs - data_res
    r0nl = sum(dprod(data_obs2,data_obs2))
    write(0,*) "Nonlinear residual norm", real(r0nl)
    write(0,*) " "

    vel1 = vel0
    b0 = 1./(vel1*vel1)
    p0 = 0.

    do iter2=1,niter2
      call model_update(vel1)
      call L_lop2(.false., .false., p0, data_res)
      data_res = data_res - data_obs2
      p_res = twt*p0
      r0 = sum(dprod(data_res,data_res)) + sum(dprod(p_res,p_res))
      call L_lop2(.true., .false., p_grad, data_res)
      p_grad = p_grad + twt*p_res
      call refl_update(p0)
      call T_lop2(.true., .false., b_grad, data_res)

      res(1) = real(sum(dprod(data_res,data_res)))
      res(2) = real(sum(dprod(p_res,p_res)))
      write(0,*) "linear residuals", res
      write(0,*) "gb stats", minval(b_grad), maxval(b_grad), sum(b_grad)/size(b_grad), sqrt(sum(b_grad*b_grad))
      write(0,*) "gp stats", minval(p_grad), maxval(p_grad), sum(p_grad)/size(p_grad), sqrt(sum(p_grad*p_grad))
      write(0,*) " "

      b_grad = b_grad + p_grad(:,:,1-otau)
      p_grad(:,:,1-otau) = b_grad
      call separate_smooth(b_grad, b_grad)
      do itau=1,ntau
        call separate_rough(p_grad(:,:,itau), p_grad(:,:,itau))
      end do
      write(0,*) "gb2 stats", minval(b_grad), maxval(b_grad), sum(b_grad)/size(b_grad), sqrt(sum(b_grad*b_grad))
      write(0,*) "gp2 stats", minval(p_grad), maxval(p_grad), sum(p_grad)/size(p_grad), sqrt(sum(p_grad*p_grad))
      write(0,*) " "

      do itry=1,2*ntry

        if(itry .eq. ntry .or. step_size .eq. 0.) then
          step_size = sqrt(sum(b0*b0)) / sqrt(sum(p_grad*p_grad)) / 1000.
          write(0,*) "Step size is resetted to ", step_size
        endif

        p1 = p0 - p_grad * step_size * 0.5
        b1 = b0 - b_grad * step_size * 0.5
        vel1 = 1./sqrt(b1)
        where(vel1<lb) vel1=lb
        where(vel1>ub) vel1=ub
        call model_update(vel1)
        call L_lop2(.false., .false., p1, data_res)
        data_res = data_res - data_obs2
        p_res = twt*p1
        r1 = sum(dprod(data_res,data_res)) + sum(dprod(p_res,p_res))

        p1 = p0 - p_grad * step_size * 1.0
        b1 = b0 - b_grad * step_size * 1.0
        vel1 = 1./sqrt(b1)
        where(vel1<lb) vel1=lb
        where(vel1>ub) vel1=ub
        call model_update(vel1)
        call L_lop2(.false., .false., p1, data_res)
        data_res = data_res - data_obs2
        p_res = twt*p1
        r2 = sum(dprod(data_res,data_res)) + sum(dprod(p_res,p_res))

        p1 = p0 - p_grad * step_size * 1.5
        b1 = b0 - b_grad * step_size * 1.5
        vel1 = 1./sqrt(b1)
        where(vel1<lb) vel1=lb
        where(vel1>ub) vel1=ub
        call model_update(vel1)
        call L_lop2(.false., .false., p1, data_res)
        data_res = data_res - data_obs2
        p_res = twt*p1
        r3 = sum(dprod(data_res,data_res)) + sum(dprod(p_res,p_res))

        write(0,*) "residuals =", real(r0), real(r1), real(r2), real(r3)
        call stepper(r0, r1, r2, r3, step_size, stat)

        if(stat .eq. 0) exit

      end do ! end of trying loop

      if(stat .ne. 0) then
        write(0,*) "trying iterations reached limit without finding a proper step size, will exit the program"
        exit
      endif

      p0 = p0 - p_grad * step_size * 0.5
      b0 = b0 - b_grad * step_size * 0.5
      vel1 = 1./sqrt(b0)
      where(vel1<lb) vel1=lb
      where(vel1>ub) vel1=ub
    end do

    vel0 = 1./sqrt(b0+p0(:,:,1-otau))
    where(vel0<lb) vel0=lb
    where(vel0>ub) vel0=ub

  end do

  call sep_write(vel0)

!---------------------------------------------------------------------------
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  write(0,*) "DONE!"
!---------------------------------------------------------------------------

  call sep_close()

end program
