#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cudaErrors.cu"

#define dloc(z,x) ((z)+(x)*n1_gpu)

#define BLOCK_SIZE 16
#define FAT 5
#define COEFFS_SIZE 11
#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4
#define CZ3 5
#define CX3 6
#define CZ4 7
#define CX4 8
#define CZ5 9
#define CX5 10

__constant__ int n1_gpu, n2_gpu, nt_gpu, wfld_j_gpu;
__constant__ int bc1_gpu, bc2_gpu, bc3_gpu, bc4_gpu;
__constant__ float *p0_gpu, *p1_gpu, *p1s_gpu;
__constant__ float *wfld1_gpu, *wfld2_gpu, *wfld3_gpu, *wfld4_gpu;
__constant__ float coeffs[COEFFS_SIZE];
__constant__ cudaArray *v1_gpu, *v2_gpu, *v3_gpu, *v4_gpu;
texture <float, 2, cudaReadModeElementType> v1_ref, v2_ref, v3_ref, v4_ref;

__global__ void wave_for_kernel(float *p1, float *p2){
  __shared__ float temp[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloc(ig,jg);                  // linear memroy address

  temp[jl][il] = (ig<n1_gpu && jg<n2_gpu) ? p1[dloc(ig,jg)] : 0.0;
  if(threadIdx.x<FAT) temp[jl][il-FAT] = (ig-FAT>=0 && ig-FAT<n1_gpu && jg<n2_gpu) ? p1[dloc(ig-FAT,jg)] : 0.0;
  if(threadIdx.x>=BLOCK_SIZE-FAT) temp[jl][il+FAT] = (ig+FAT<n1_gpu && jg<n2_gpu) ? p1[dloc(ig+FAT,jg)] : 0.0;
  if(threadIdx.y<FAT) temp[jl-FAT][il] = (jg-FAT>=0 && jg-FAT<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg-FAT)] : 0.0;
  if(threadIdx.y>=BLOCK_SIZE-FAT) temp[jl+FAT][il] = (jg+FAT<n2_gpu && ig<n1_gpu) ? p1[dloc(ig,jg+FAT)] : 0.0;
  __syncthreads();

  if(ig>=n1_gpu || jg>=n2_gpu) return;

  p2[addr] += tex2D(v3_ref,ig,jg)*temp[jl][il] + tex2D(v1_ref,ig,jg)*(
      coeffs[CX5]*(temp[jl-5][il]+temp[jl+5][il])
      +coeffs[CZ5]*(temp[jl][il-5]+temp[jl][il+5])
      +coeffs[CX4]*(temp[jl-4][il]+temp[jl+4][il])
      +coeffs[CZ4]*(temp[jl][il-4]+temp[jl][il+4])
      +coeffs[CX3]*(temp[jl-3][il]+temp[jl+3][il])
      +coeffs[CZ3]*(temp[jl][il-3]+temp[jl][il+3])
      +coeffs[CX2]*(temp[jl-2][il]+temp[jl+2][il])
      +coeffs[CZ2]*(temp[jl][il-2]+temp[jl][il+2])
      +coeffs[CX1]*(temp[jl-1][il]+temp[jl+1][il])
      +coeffs[CZ1]*(temp[jl][il-1]+temp[jl][il+1])
      +coeffs[C0]*temp[jl][il]);

  if(ig<bc1_gpu) p2[addr] += tex2D(v2_ref,ig,jg)*(temp[jl][il]-temp[jl][il+1]);
  if(ig>=n1_gpu-bc3_gpu) p2[addr] += tex2D(v2_ref,ig,jg)*(temp[jl][il]-temp[jl][il-1]);
  if(jg<bc2_gpu) p2[addr] += tex2D(v2_ref,ig,jg)*(temp[jl][il]-temp[jl+1][il]);
  if(jg>=n2_gpu-bc4_gpu) p2[addr] += tex2D(v2_ref,ig,jg)*(temp[jl][il]-temp[jl-1][il]);
}

__global__ void wave_adj_kernel(float *p1, float *p1s, float *p2){
  __shared__ float temp[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
  register int ig=blockIdx.x*blockDim.x+threadIdx.x;   // global z coordinates
  register int jg=blockIdx.y*blockDim.y+threadIdx.y;   // global x coordinates
  register int il=FAT+threadIdx.x;               // thread-local z coordinate
  register int jl=FAT+threadIdx.y;               // thread-local x coordinate
  register int addr=dloc(ig,jg);                  // linear memroy address

  temp[jl][il] = (ig<n1_gpu && jg<n2_gpu) ? p1s[dloc(ig,jg)] : 0.0;
  if(threadIdx.x<FAT) temp[jl][il-FAT] = (ig-FAT>=0 && ig-FAT<n1_gpu && jg<n2_gpu) ? p1s[dloc(ig-FAT,jg)] : 0.0;
  if(threadIdx.x>=BLOCK_SIZE-FAT) temp[jl][il+FAT] = (ig+FAT<n1_gpu && jg<n2_gpu) ? p1s[dloc(ig+FAT,jg)] : 0.0;
  if(threadIdx.y<FAT) temp[jl-FAT][il] = (jg-FAT>=0 && jg-FAT<n2_gpu && ig<n1_gpu) ? p1s[dloc(ig,jg-FAT)] : 0.0;
  if(threadIdx.y>=BLOCK_SIZE-FAT) temp[jl+FAT][il] = (jg+FAT<n2_gpu && ig<n1_gpu) ? p1s[dloc(ig,jg+FAT)] : 0.0;
  __syncthreads();

  if(ig>=n1_gpu || jg>=n2_gpu) return;

  p2[addr] += tex2D(v3_ref,ig,jg)*p1[addr]+
      coeffs[CX5]*(temp[jl-5][il]+temp[jl+5][il])
      +coeffs[CZ5]*(temp[jl][il-5]+temp[jl][il+5])
      +coeffs[CX4]*(temp[jl-4][il]+temp[jl+4][il])
      +coeffs[CZ4]*(temp[jl][il-4]+temp[jl][il+4])
      +coeffs[CX3]*(temp[jl-3][il]+temp[jl+3][il])
      +coeffs[CZ3]*(temp[jl][il-3]+temp[jl][il+3])
      +coeffs[CX2]*(temp[jl-2][il]+temp[jl+2][il])
      +coeffs[CZ2]*(temp[jl][il-2]+temp[jl][il+2])
      +coeffs[CX1]*(temp[jl-1][il]+temp[jl+1][il])
      +coeffs[CZ1]*(temp[jl][il-1]+temp[jl][il+1])
      +coeffs[C0]*temp[jl][il];
}

__global__ void scale_kernel(float *p1, float *p1s)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  p1s[dloc(i,j)] = tex2D(v1_ref,i,j)*p1[dloc(i,j)];
}

__global__ void bc_scale_kernel(float *p1, float *p1s)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  p1s[dloc(i,j)] = tex2D(v2_ref,i,j)*p1[dloc(i,j)];
}

__global__ void bc_kernel(float *p1s, float *p2)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  if(i==0) p2[dloc(i,j)] += p1s[dloc(i,j)];
  if(i>0 && i<bc1_gpu) p2[dloc(i,j)] += p1s[dloc(i,j)]-p1s[dloc(i-1,j)];
  if(i==bc1_gpu) p2[dloc(i,j)] -= p1s[dloc(i-1,j)];
  if(i==n1_gpu-bc3_gpu-1) p2[dloc(i,j)] -= p1s[dloc(i+1,j)];
  if(i<n1_gpu-1 && i>=n1_gpu-bc3_gpu) p2[dloc(i,j)] += p1s[dloc(i,j)]-p1s[dloc(i+1,j)];
  if(i==n1_gpu-1) p2[dloc(i,j)] += p1s[dloc(i,j)];
  if(j==0) p2[dloc(i,j)] += p1s[dloc(i,j)];
  if(j>0 && j<bc2_gpu) p2[dloc(i,j)] += p1s[dloc(i,j)]-p1s[dloc(i,j-1)];
  if(j==bc2_gpu) p2[dloc(i,j)] -= p1s[dloc(i,j-1)];
  if(j==n2_gpu-bc4_gpu-1) p2[dloc(i,j)] -= p1s[dloc(i,j+1)];
  if(j<n2_gpu-1 && j>=n2_gpu-bc4_gpu) p2[dloc(i,j)] += p1s[dloc(i,j)]-p1s[dloc(i,j+1)];
  if(j==n2_gpu-1) p2[dloc(i,j)] += p1s[dloc(i,j)];
}

__global__ void inject_wfld_kernel(float *p0, float *wfld1, float *wfld2, float lwt1, float lwt2, float *p2)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  p2[dloc(i,j)] = lwt1*wfld1[i+j*n1_gpu] + lwt2*wfld2[i+j*n1_gpu] + tex2D(v4_ref,i,j)*p0[dloc(i,j)];
}

__global__ void collect_wfld_kernel(float *p, float lwt1, float lwt2, float *wfld1, float *wfld2)
{
  register int i=blockIdx.x*blockDim.x+threadIdx.x;
  register int j=blockIdx.y*blockDim.y+threadIdx.y;
  if(i>=n1_gpu || j>=n2_gpu) return;
  wfld1[i+j*n1_gpu] += lwt1*p[dloc(i,j)];
  wfld2[i+j*n1_gpu] += lwt2*p[dloc(i,j)];
}

extern "C" void gpu_setdevice(int *gpu_id)
{
  cudaSetDevice(*gpu_id);
}

extern "C" void gpu_init(int *n1_in, int *n2_in, int *nt_in, int *wfld_j_in, int *bc1_in, int *bc2_in, int *bc3_in, int *bc4_in, float *d1_in, float *d2_in)
{
  int n1, n2, nt, wfld_j, bc1, bc2, bc3, bc4;
  float d1, d2;
  float *p0, *p1, *p1s, *wfld1, *wfld2, *wfld3, *wfld4;
  long int memsize;
  cudaArray *v1, *v2, *v3, *v4;

  n1 = (*n1_in);
  n2 = (*n2_in);
  nt = (*nt_in);
  wfld_j = (*wfld_j_in);
  bc1 = (*bc1_in);
  bc2 = (*bc2_in);
  bc3 = (*bc3_in);
  bc4 = (*bc4_in);
  d1 = (*d1_in);
  d2 = (*d2_in);

  cudaMemcpyToSymbol(n1_gpu,  &n1,  sizeof(int));
  cudaMemcpyToSymbol(n2_gpu,  &n2,  sizeof(int));
  cudaMemcpyToSymbol(nt_gpu,  &nt,  sizeof(int));
  cudaMemcpyToSymbol(wfld_j_gpu, &wfld_j, sizeof(int));
  cudaMemcpyToSymbol(bc1_gpu, &bc1, sizeof(int));
  cudaMemcpyToSymbol(bc2_gpu, &bc2, sizeof(int));
  cudaMemcpyToSymbol(bc3_gpu, &bc3, sizeof(int));
  cudaMemcpyToSymbol(bc4_gpu, &bc4, sizeof(int));

  float coeffs_cpu[COEFFS_SIZE];
  coeffs_cpu[CZ1]= 3.333333333/(2.*d1*d1);
  coeffs_cpu[CX1]= 3.333333333/(2.*d2*d2);
  coeffs_cpu[CZ2]=-.4761904762/(2.*d1*d1);
  coeffs_cpu[CX2]=-.4761904762/(2.*d2*d2);
  coeffs_cpu[CZ3]= .0793650794/(2.*d1*d1);
  coeffs_cpu[CX3]= .0793650794/(2.*d2*d2);
  coeffs_cpu[CZ4]=-.0099206349/(2.*d1*d1);
  coeffs_cpu[CX4]=-.0099206349/(2.*d2*d2);
  coeffs_cpu[CZ5]= .0006349206/(2.*d1*d1);
  coeffs_cpu[CX5]= .0006349206/(2.*d2*d2);
  coeffs_cpu[C0]=coeffs_cpu[CZ5]+coeffs_cpu[CX5];
  coeffs_cpu[C0]+=coeffs_cpu[CZ4]+coeffs_cpu[CX4];
  coeffs_cpu[C0]+=coeffs_cpu[CZ3]+coeffs_cpu[CX3];
  coeffs_cpu[C0]+=coeffs_cpu[CZ2]+coeffs_cpu[CX2];
  coeffs_cpu[C0]+=coeffs_cpu[CZ1]+coeffs_cpu[CX1];
  coeffs_cpu[C0]=-2.*coeffs_cpu[C0];
  cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float));

  cudaMalloc( (void**) &p0,       n1*n2*sizeof(float));
  cudaMalloc( (void**) &p1,       n1*n2*sizeof(float));
  cudaMalloc( (void**) &p1s,      n1*n2*sizeof(float));
  cudaMalloc( (void**) &wfld1,    n1*n2*sizeof(float));
  cudaMalloc( (void**) &wfld2,    n1*n2*sizeof(float));
  cudaMalloc( (void**) &wfld3,    n1*n2*sizeof(float));
  cudaMalloc( (void**) &wfld4,    n1*n2*sizeof(float));

  cudaChannelFormatDesc cf;
  cf = cudaCreateChannelDesc<float>();
  cudaMallocArray(&v1, &cf, n1, n2);
  cudaMallocArray(&v2, &cf, n1, n2);
  cudaMallocArray(&v3,  &cf, n1, n2);
  cudaMallocArray(&v4,   &cf, n1, n2);

  memsize = (long int)n1*n2*11;

  int tempgpuid;
  cudaGetDevice(&tempgpuid);
  fprintf(stderr, " GPU %d memory allocated %f MB (%f GB) \n", tempgpuid, memsize*4./1024/1024, memsize*4./1024/1024/1024);

  cudaMemcpyToSymbol(p0_gpu, &p0, sizeof(float*));
  cudaMemcpyToSymbol(p1_gpu, &p1, sizeof(float*));
  cudaMemcpyToSymbol(p1s_gpu, &p1s, sizeof(float*));
  cudaMemcpyToSymbol(wfld1_gpu, &wfld1, sizeof(float*));
  cudaMemcpyToSymbol(wfld2_gpu, &wfld2, sizeof(float*));
  cudaMemcpyToSymbol(wfld3_gpu, &wfld3, sizeof(float*));
  cudaMemcpyToSymbol(wfld4_gpu, &wfld4, sizeof(float*));
  cudaMemcpyToSymbol(v1_gpu, &v1, sizeof(float*));
  cudaMemcpyToSymbol(v2_gpu, &v2, sizeof(float*));
  cudaMemcpyToSymbol(v3_gpu, &v3, sizeof(float*));
  cudaMemcpyToSymbol(v4_gpu, &v4, sizeof(float*));
}

extern "C" void gpu_model_update(float *v1_cpu, float *v2_cpu, float *v3_cpu, float *v4_cpu)
{
  int n1, n2;
  cudaArray *v1, *v2, *v3, *v4;

  cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int));

  cudaMemcpyFromSymbol(&v1, v1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&v2, v2_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&v3, v3_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&v4, v4_gpu, sizeof(float*));

  cudaMemcpyToArray(v1, 0,0, v1_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpyToArray(v2, 0,0, v2_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpyToArray(v3, 0,0, v3_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpyToArray(v4, 0,0, v4_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);

  cudaBindTextureToArray(v1_ref, v1);
  cudaBindTextureToArray(v2_ref, v2);
  cudaBindTextureToArray(v3_ref, v3);
  cudaBindTextureToArray(v4_ref, v4);
}

extern "C" void gpu_prop(float *wfld1_cpu, float *wfld2_cpu)
{
  int n1, n2, nt, wfld_j;
  float lwt1, lwt2, c;
  float *p0, *p1, *wfld1, *wfld2, *wfld3, *wfld4, *temp;

  cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld1, wfld1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld2, wfld2_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld3, wfld3_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld4, wfld4_gpu, sizeof(float*));

  int nblocksx=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);

  cudaMemset(p0,    0,n1*n2*sizeof(float));
  cudaMemset(p1,    0,n1*n2*sizeof(float));
  cudaMemset(wfld1, 0,n1*n2*sizeof(float));
  cudaMemset(wfld2, 0,n1*n2*sizeof(float));
  cudaMemset(wfld3, 0,n1*n2*sizeof(float));
  cudaMemset(wfld4, 0,n1*n2*sizeof(float));

  cudaMemcpy(wfld1, wfld1_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(wfld3, wfld2_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice);

  c = 1./sqrt(1.*wfld_j);
  lwt1 = c;
  lwt2 = 0.;
  kernel_exec(inject_wfld_kernel<<<dimGrid, dimBlock>>>(p0, wfld1, wfld2, lwt1, lwt2, p1));
  kernel_exec(collect_wfld_kernel<<<dimGrid, dimBlock>>>(p1, lwt1, lwt2, wfld3, wfld4));

  for(int it=0; it<nt-1; it++)
  {
    cudaMemcpy(wfld2, &wfld1_cpu[(long int)n1*n2*(it+1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);
    cudaMemcpy(wfld4, &wfld2_cpu[(long int)n1*n2*(it+1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);
    for(int iw=0; iw<wfld_j; iw++)
    {
      lwt2 = c*(iw+1.)/wfld_j;
      lwt1 = c - lwt2;
      kernel_exec(inject_wfld_kernel<<<dimGrid, dimBlock>>>(p0, wfld1, wfld2, lwt1, lwt2, p0));
      kernel_exec(wave_for_kernel<<<dimGrid, dimBlock>>>(p1, p0));
      kernel_exec(collect_wfld_kernel<<<dimGrid, dimBlock>>>(p0, lwt1, lwt2, wfld3, wfld4));
      temp = p0;
      p0 = p1;
      p1 = temp;
    }
    cudaMemcpy(&wfld2_cpu[(long int)n1*n2*it], wfld3, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
    temp = wfld1;
    wfld1 = wfld2;
    wfld2 = temp;
    temp = wfld3;
    wfld3 = wfld4;
    wfld4 = temp;
  }
    cudaMemcpy(&wfld2_cpu[(long int)n1*n2*(nt-1)], wfld3, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
}

extern "C" void gpu_prop_adj(float *wfld1_cpu, float *wfld2_cpu)
{
  int n1, n2, nt, wfld_j;
  float lwt1, lwt2, c;
  float *p0, *p1, *p1s, *wfld1, *wfld2, *wfld3, *wfld4, *temp;

  cudaMemcpyFromSymbol(&n1, n1_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&n2, n2_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&wfld_j, wfld_j_gpu, sizeof(int));
  cudaMemcpyFromSymbol(&p0, p0_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&p1, p1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&p1s, p1s_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld1, wfld1_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld2, wfld2_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld3, wfld3_gpu, sizeof(float*));
  cudaMemcpyFromSymbol(&wfld4, wfld4_gpu, sizeof(float*));

  int nblocksx=ceil(1.*n1/BLOCK_SIZE);
  int nblocksy=ceil(1.*n2/BLOCK_SIZE);
  dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
  dim3 dimGrid(nblocksx,nblocksy);

  cudaMemset(p0,    0,n1*n2*sizeof(float));
  cudaMemset(p1,    0,n1*n2*sizeof(float));
  cudaMemset(p1s,    0,n1*n2*sizeof(float));
  cudaMemset(wfld1, 0,n1*n2*sizeof(float));
  cudaMemset(wfld2, 0,n1*n2*sizeof(float));
  cudaMemset(wfld3, 0,n1*n2*sizeof(float));
  cudaMemset(wfld4, 0,n1*n2*sizeof(float));

  cudaMemcpy(wfld1, &wfld1_cpu[(long int)n1*n2*(nt-1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(wfld3, &wfld2_cpu[(long int)n1*n2*(nt-1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);

  c = 1./sqrt(1.*wfld_j);
  lwt1 = c;
  lwt2 = 0.;
  kernel_exec(inject_wfld_kernel<<<dimGrid, dimBlock>>>(p0, wfld1, wfld2, lwt1, lwt2, p1));
  kernel_exec(collect_wfld_kernel<<<dimGrid, dimBlock>>>(p1, lwt1, lwt2, wfld3, wfld4));

  for(int it=nt-1; it>=1; it--)
  {
    cudaMemcpy(wfld2, &wfld1_cpu[(long int)n1*n2*(it-1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);
    cudaMemcpy(wfld4, &wfld2_cpu[(long int)n1*n2*(it-1)], n1*n2*sizeof(float), cudaMemcpyHostToDevice);
    for(int iw=0; iw<wfld_j; iw++)
    {
      lwt2 = c*(iw+1.)/wfld_j;
      lwt1 = c - lwt2;
      kernel_exec(inject_wfld_kernel<<<dimGrid, dimBlock>>>(p0, wfld1, wfld2, lwt1, lwt2, p0));
      kernel_exec(scale_kernel<<<dimGrid, dimBlock>>>(p1, p1s));
      kernel_exec(wave_adj_kernel<<<dimGrid, dimBlock>>>(p1, p1s, p0));
      kernel_exec(bc_scale_kernel<<<dimGrid, dimBlock>>>(p1, p1s));
      kernel_exec(bc_kernel<<<dimGrid, dimBlock>>>(p1s, p0));
      kernel_exec(collect_wfld_kernel<<<dimGrid, dimBlock>>>(p0, lwt1, lwt2, wfld3, wfld4));
      temp = p0;
      p0 = p1;
      p1 = temp;
    }
    cudaMemcpy(&wfld2_cpu[(long int)n1*n2*it], wfld3, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
    temp = wfld1;
    wfld1 = wfld2;
    wfld2 = temp;
    temp = wfld3;
    wfld3 = wfld4;
    wfld4 = temp;
  }
    cudaMemcpy(wfld2_cpu, wfld3, n1*n2*sizeof(float), cudaMemcpyDeviceToHost);
}
